#include "buffer.h"
#include "bailout.h"
#include "uogetopt.h"
#include "out.h"
#include "scan.h"
#include "divedata.h"
#include "mosquitotools.h"
#include "stdio.h"
#include "scan__float.h"
#include "open.h"
#include "readwrite.h"
#include "error.h"
#include "close.h"
#include "getln.h"
#include "str.h"
#include "caltime.h"

static int o_force;
static unsigned long o_surface_pause;

static uogetopt2 opts[]={
{  'f',"force",uogo_flag,UOGO_NOARG,&o_force,1,
 "Do join dives even if they don't match.",
 "This allows to join air and nitrox dives, for example. The result is "
 "probably not desirable.\n",0},
{ 's',"surface-pause",do_timelength_opt,0,&o_surface_pause,0,
 "Assume a surface pause of TIME.",
 "TIME is an integer value defaulting to seconds. A time unit may be "
 "appended directly to the number:\n"
 " * seconds (default)\n"
 " * minutes\n"
 " * hours\n"
 " * days\n"
 "When this option is not given join-dive-profiles will try to determine "
 "the surface pause from the dive profiles.","TIME"},
{  0,"limitations",uogo_print_help,UOGO_NOARG|UOGO_EXIT|UOGO_NOLHD,
  0,0,
  "Known limitations.",
  "This program ignores the configuration data of the second dive. It mixes "
  "the dive information of the two dives, using the following information "
  "from the first dive:\n"
  " * surface-interval\n"
  " * temperature-at-start\n"
  " * dive-start-date\n"
  " * total-dive-number (a user supplied information)\n"
  "using these from the second dive:\n"
  " * OLF-value\n"
  " * temperature-at-end\n"
  "and these from the first dive, if available, otherwise from the second "
  "dive:\n"
  " * dive-mode\n"
  " * dive-no (the dive number in the repetition group)\n"
  " * recording-interval\n"
  " * altitude-setting\n"
  " * personal-setting\n"
  " * oxygen-percentage\n"
  " * place (a user supplied value)\n"
  " * equipment (a user supplied value)\n"
  " * buddies (a user supplied value)\n"
  " * buddies (a user supplied value)\n"
  "The `temperature-at-max-depth' is calculated from the lowest temperature of "
  "both temperature-of-max-depth values, the temperature-at-start of the "
  "seconds dive and the temperature-at-end of the first dive.\n"
  "The `max-depth' is the maximum depth of the two dives, and the "
  "`total-seconds' is the sum of the length of the two dives and the gap "
  "between them.\n"
  "The comments for both dives are preserved in order.\n\n"
  "This is certainly less than good in some circumstances, but there is "
  "no way a program will get this right in all circumstances. Sorry."
  ,0},

  {0,"",uogo_include,0,general_opts, 0,0,0,0},
  {0,0,0,0,0,0,0,0,0}
  };
static uogetopt_env optenv={
  0,PACKAGE,VERSION,
    "usage: joing-dive-profiles DIVE1 DIVE2\n",
"Concatenate two dive profiles.",
"join-dive-profiles creates a dive profile from two dive profiles.\n"
"The new profile is written to the standard output.\n\n"
"Be sure to read the manual page to understand this programs limitations.\n"
,
COMMON_BUGREPORT_INFO,
    2,0,
    0,0,
    uogetopt_out,
    opts
};

static int check_num;
static struct tai check_end1;
static struct tai check_start2;
static unsigned long dive1endtime;
static unsigned long surface_interval2;
static strhash h1;
static strhash h2;
static void 
check1(struct dive *d)
{
  strhash *h;
  char *key,*data;
  uint32 kl,dl;

  if (check_num==1) {
    struct tai x;
    dive1endtime=d->data.ix[d->data.len-1].endtime;
    tai_uint(&x,dive1endtime);
    if (d->start_time.x)
      tai_add(&check_end1,&d->start_time,&x);
    h=&h1;
  }
  else {
    unsigned long ul;
    const char *s;
    unsigned int l;
    check_start2=d->start_time;
    surface_interval2=0;
    s=gv(&d->h_dive,"surface-interval-hours");
    if (s) {
      l=scan_ulong(s,&ul);
      if (l) {
	surface_interval2+=ul*3600;
      }
    }
    s=gv(&d->h_dive,"surface-interval-minutes");
    if (s) {
      l=scan_ulong(s,&ul);
      if (l) {
	surface_interval2+=ul*60;
      }
    }
    s=gv(&d->h_dive,"surface-interval-seconds");
    if (s) {
      l=scan_ulong(s,&ul);
      if (l) {
	surface_interval2+=ul;
      }
    }
    h=&h2;
  }
  if (-1==strhash_create(h,1,32,strhash_hash)) oom();
  strhash_walkstart(&d->h_dive);
  while (strhash_walk(&d->h_dive,&key,&kl,&data,&dl))
    if (-1==strhash_enter(h,1,key,kl,1,data,dl)) oom();
}

int main(int argc, char **argv)
{
  uogetopt2 *o;
  int fd;
  buffer b;
  char bspace[BUFFER_INSIZE];
  stralloc sa=STRALLOC_INIT;
  unsigned long offset;
  unsigned long gap;
  unsigned long total_seconds=0; /* keep gcc quiet */
  int dives_match;

  flag_bailout_fatal_begin=3;
  uogo_posixmode=1;

  bailout_progname(argv[0]);
  o=uogetopt_join(opts);;
  if (!o) oom();
  optenv.program=flag_bailout_log_name;
  optenv.opts=o;
  uogetopt_parse (&optenv, &argc, argv);
  if (argc!=3)
    xbailout(100,0,"need two dive profiles.",0,0,0);
  check_num=1;
  format_handler(argv[1],check1);
  check_num=2;
  format_handler(argv[2],check1);
  
  if (o_surface_pause) {
    offset=o_surface_pause+dive1endtime;
    gap=o_surface_pause;
  } else if (surface_interval2) {
    gap=surface_interval2;
    offset=gap+dive1endtime;
  } else {
    if (!tai_less(&check_end1,&check_start2)) {
      struct caltime ct;
      char buf[128];
      unsigned int l;
      unsigned long end,start;
      end=tai_approx(&check_end1);
      start=tai_approx(&check_start2);
      if (!end) 
	xbailout(100,0,argv[1], " does not contain time information.",0,0); 
      if (!start) 
	xbailout(100,0,argv[2], " does not contain time information.",0,0); 
      /* dive start time doesn't contain seconds */
      if (start>=end+60) {
	caltime_utc(&ct,&check_end1,0,0);
	l=caltime_fmt(buf,&ct);
	buf[l]=0;
	warning(0,"dive ",argv[1], " ends ",buf);

	caltime_utc(&ct,&check_start2,0,0);
	l=caltime_fmt(buf,&ct);
	buf[l]=0;
	warning(0,"dive ",argv[2], " starts ",buf);

	xbailout(100,0,"bad dive order",0,0,0);
      }
      gap=30; /* assumption */
      offset=gap+dive1endtime;
    }
    else {
      struct tai x;
      tai_sub(&x,&check_start2,&check_end1);
      gap=tai_approx(&x);
      offset=gap+dive1endtime;
    }
  }

#define FORCE_EQUAL(SEARCH) \
do { \
  static stralloc sa1; \
  const char *x; \
  x=gv(&h1,SEARCH); \
  if (!x) { \
    x=gv(&h2,SEARCH); \
    if (!x) break; \
    warning(0,SEARCH," is not in ",argv[1],0); \
    dives_match=0; \
    break; \
  } \
  if (!stralloc_copys(&sa1,x)) oom(); \
  if (!stralloc_0(&sa1)) oom(); \
  x=gv(&h2,SEARCH); \
  if (!x) { \
    warning(0,SEARCH," is not in ",argv[2],0); \
    dives_match=0; \
    break; \
  } \
  if (str_diff(x,sa1.s)) { \
    warning(0,"different values for ",SEARCH," in ",0); \
    warning(0,"  ",argv[1],": ",sa1.s); \
    warning(0,"  ",argv[2],": ",x); \
    dives_match=0; \
  } \
} while(0) 

  dives_match=1;
  FORCE_EQUAL("dive-mode");
  FORCE_EQUAL("oxygen-percentage");

  if (!dives_match && !o_force) xbailout(100,0,"dives do not match",0,0,0);

  fd=open_read(argv[1]);
  if (-1==fd) xbailout(111,errno,"failed to open ",argv[1],0,0);
  buffer_init(&b,(buffer_op_t) read, fd, bspace,sizeof(bspace));

  /* copy conf */
  while (1) {
    int gotlf;
    sa.len=0;
    if (-1==getln(&b,&sa,&gotlf,'\n'))
      xbailout(111,errno,"failed to read ",argv[1],0,0);
    if (!sa.len) break;
    if (!gotlf)
      xbailout(100,0,"unterminated line in ",argv[1],0,0);
    sa.s[sa.len-1]=0; /* \n -> \0 */
    outs(sa.s);
    outs("\n");
    if (str_start(sa.s,"*dive")) 
      break;
  }
  /* copy all except the colon lines, stop before '#' */
  while (1) {
    int gotlf;
    sa.len=0;
    if (-1==getln(&b,&sa,&gotlf,'\n'))
      xbailout(111,errno,"failed to read ",argv[1],0,0);
    if (!sa.len) break;
    if (!gotlf)
      xbailout(100,0,"unterminated line in ",argv[1],0,0);
    sa.s[sa.len-1]=0; /* \n -> \0 */
    if (sa.s[0]=='#') 
      break;
    if (sa.s[0]!=':') {
      outs(sa.s);
      outs("\n");
    }
  }
  close(fd);

  /* now fill the gap */
  if (gap>20) {
    unsigned long t;
    for (t=20;t<gap;t+=20) {
      outs("D");
      outu(t+dive1endtime);
      outs(" 0 dummy\n");
    }
  }

  fd=open_read(argv[2]);
  if (-1==fd) xbailout(111,errno,"failed to open ",argv[2],0,0);
  buffer_init(&b,(buffer_op_t) read, fd, bspace,sizeof(bspace));

  /* skip conf */
  while (1) {
    int gotlf;
    sa.len=0;
    if (-1==getln(&b,&sa,&gotlf,'\n'))
      xbailout(111,errno,"failed to read ",argv[2],0,0);
    if (!sa.len) break;
    if (!gotlf)
      xbailout(100,0,"unterminated line in ",argv[2],0,0);
    sa.s[sa.len-1]=0; /* \n -> \0 */
    if (str_start(sa.s,"*dive")) 
      break;
  }
  /* copy R lines and the modified D lines only */
  while (1) {
    int gotlf;
    sa.len=0;
    if (-1==getln(&b,&sa,&gotlf,'\n'))
      xbailout(111,errno,"failed to read ",argv[2],0,0);
    if (!sa.len) break;
    if (!gotlf)
      xbailout(100,0,"unterminated line in ",argv[2],0,0);
    sa.s[sa.len-1]=0; /* \n -> \0 */
    if (sa.s[0]=='#') 
      break;
    if (sa.s[0]=='R') {
      outs(sa.s);
      outs("\n");
      continue;
    }
    if (sa.s[0]!='D')
      continue;
    else
    {
    /* Dtime depth anything else */
      unsigned int l;
      unsigned long ul;
      l=scan_ulong(sa.s+1,&ul);
      if (!l || sa.s[l+1]!=' ')
	xbailout(100,0,"bad format in ",argv[2], " in line ",sa.s);
      outs("D");
      outu(ul+offset);
      outs(sa.s+l+1);
      outs("\n");
      total_seconds=gap+dive1endtime+ul;
    }
  }
  close(fd);

#define PREFER_FIRST(SEARCH) \
  do { \
    const char *tmp; \
    tmp=gv(&h1,SEARCH); \
    if (!tmp)  \
      tmp=gv(&h2,SEARCH); \
    if (tmp) { \
      outs(":"); outs(SEARCH); outs(" "); outs(tmp); outs("\n"); \
    } \
  } while(0)
#define ONLY_FIRST(SEARCH)  \
  do { \
    const char *tmp; \
    tmp=gv(&h1,SEARCH); \
    if (tmp) { \
      outs(":"); outs(SEARCH); outs(" "); outs(tmp); outs("\n"); \
    } \
  } while(0)
#define ONLY_SECOND(SEARCH)  \
  do { \
    const char *tmp; \
    tmp=gv(&h2,SEARCH); \
    if (tmp) { \
      outs(":"); outs(SEARCH); outs(" "); outs(tmp); outs("\n"); \
    } \
  } while(0)

  /* now handle the colon stuff. */
  PREFER_FIRST("dive-mode");
  ONLY_FIRST("surface-interval-minutes");
  ONLY_FIRST("surface-interval-hours");
  PREFER_FIRST("dive-no");
  PREFER_FIRST("recording-interval");
  PREFER_FIRST("altitude-setting");
  PREFER_FIRST("personal-setting");
  PREFER_FIRST("oxygen-percentage");
  ONLY_SECOND("OLF-value");
  ONLY_FIRST("temperature-at-start");
  /* Oh, how nice */
  {
    unsigned long min=100; /* boiling water */
    unsigned long ul;
    const char *s;
    s=gv(&h1,"temperature-at-max-depth");
    if (s && scan_ulong(s,&ul))
      if (ul<min) min=ul;
    s=gv(&h1,"temperature-at-end");
    if (s && scan_ulong(s,&ul))
      if (ul<min) min=ul;
    s=gv(&h2,"temperature-at-start");
    if (s && scan_ulong(s,&ul))
      if (ul<min) min=ul;
    s=gv(&h2,"temperature-at-max-depth");
    if (s && scan_ulong(s,&ul))
      if (ul<min) min=ul;
    if (min!=100) {
      outs(":temperature-at-max-depth ");
      outu(min);
      outs("\n");
    }
  }
  ONLY_SECOND("temperature-at-end");
  ONLY_FIRST("dive-start-date");
  /* Oh, how nice */
  {
    float max=0;
    float f;
    const char *s;
    s=gv(&h1,"max-depth");
    if (s && scan__float(s,&f)) 
      max=f;
    s=gv(&h2,"max-depth");
    if (s && scan__float(s,&f) && f>max) 
      max=f;
    if (max>0.1) {
      outs(":max-depth ");
      outf(max,2);
      outs("\n");
    }
  }
  outs(":total-seconds ");
  outu(total_seconds);
  outs("\n");

  ONLY_FIRST("total-dive-number");
  /* handle bottle and pressure */
  {
    const char *s;
    s=gv(&h1,"bottle-volume");
    if (s) {
      static stralloc sa1;
      if (!stralloc_copys(&sa1,s)) oom();
      if (!stralloc_0(&sa1)) oom();
      s=gv(&h2,"bottle-volume");
      if (s && str_equal(s,sa1.s)) {
	outs(":"); outs("bottle-volume"); outs(" "); outs(s); outs("\n");
	ONLY_FIRST("pressure-at-begin");
	ONLY_SECOND("pressure-at-end");
      }
    }
  }
  PREFER_FIRST("place");
  PREFER_FIRST("equipment");
  PREFER_FIRST("buddies");
  /* take care of comments */
  {
    unsigned int i=0;
    unsigned int k=0;
    for (i=1;i<=2;i++) {
      unsigned int j=0;
      while (1) {
	const char *s;
	if (!stralloc_copys(&sa,"comment")) oom();
	if (!stralloc_catuint0(&sa,j++,0)) oom();
	if (!stralloc_0(&sa)) oom();
	s=gv(i==1 ? &h1 : &h2,sa.s);
	if (!s) break;
	if (s && *s) {
	  outs(":comment");
	  outu(k++);
	  outs(" ");
	  outs(s);
	  outs("\n");
	}
      }
    }
  }
  outflush();
  return 0;
}
