#include "buffer.h"
#include "getln.h"
#include "bailout.h"
#include "error.h"
#include "str.h"
#include "fmt.h"
#include "open.h"
#include "buffer.h"
#include "fmt.h"
#include "math.h"
#include "out.h"
#include "uogetopt.h"
#include "stralloc.h"
#include "divedata.h"
#include "scan.h"
#include "mosquitotools.h"
#include "zhl16.h"
#include "scan__float.h"
#include "alloc.h"
#include <unistd.h>
#include <stdio.h>
#include <time.h>

#define DEFAULT_LOAD_MARGIN 0.737 /* -L */
#define DEFAULT_DECO_FACTOR 1.00  /* -D */
#define DEFAULT_DESCEND_SPEED 20
#define DEFAULT_ASCEND_SPEED 10
#define DEFAULT_MODEL        "zhl16b"
#define DEFAULT_DECO_STEP_TIME_LIMIT 300 /* minutes */

static int           o_profile=0;
static int           o_verbose=0;
static int           o_feet=0;
static float         feetfactor=1;

static int           o_mdp=0;
static float         o_descend_speed=DEFAULT_DESCEND_SPEED;
static float         o_ascend_speed=DEFAULT_ASCEND_SPEED;
static unsigned long o_deco_step_time_limit=DEFAULT_DECO_STEP_TIME_LIMIT;
static float         o_rmv=0;
static float         o_air_pressure=1.0;
static const char   *o_zhl16_setting=DEFAULT_MODEL;
static float         o_nitrox_fraction=0.78;
static float         o_load_margin=DEFAULT_LOAD_MARGIN;
static float baseload;

static unsigned int   deco_depths_count;
static float         *deco_depths;

static void 
warn_bad_table(const char *s)
{
  warning(0,"The generated table must not be used: ",s,0,0);
  outs("*** The generated table must not be used: ");
  outs(s);
  outs("\n");
}
static void
verbose(unsigned long gt, const char *reason, int p, unsigned long x)
{
  if (!o_verbose)
    return;
  outun(gt/60,5);
  outs(": ");
  outs(reason);
  if (p) {
    outs(": ");
    outu(x);
  }
  outs("\n");
}


/*
 * Assumptions:
 * 1. the diver descends with lightning speed (safer side), unless
 *    o_descend_speed is set.
 * 2. the diver ascends with o_ascend_speed/sec.
 * 3. the calculations assume that the diver during ascend stays at the
 *    old depth for 6 seconds and then ascends 1 meter without needing 
 *    time. This should be quite safe, too.
 */
static int 
do_one_dive(float target_depth, unsigned long ground_time, 
  float *start_load, unsigned long *deco_minutes, float *barliter)
{
  /* note: all calculations done in meter, second and bar */
  float deco_depth=0;
  unsigned int gewebe;
  float depth;
  unsigned long total_time;
  float sim_depth;
  unsigned int g;
  unsigned int deco_pointer;

  float *sim_load=zhl16_alloc(&gewebe);
  float *maxload=zhl16_alloc(&gewebe);
  float *load=zhl16_alloc(&gewebe);
  if (!sim_load || !load || !maxload) oom();
  zhl16_copy(load,start_load);
  *barliter=0;

#define AIR(depth) ((o_air_pressure+depth/10)*o_rmv)

  total_time=0;
  if (o_descend_speed>0) {
    depth=0;
    while (depth<target_depth) {
      depth+=o_descend_speed/60*6;
      if (depth>target_depth)
	depth=target_depth;
      zhl16_calc_load(load,o_air_pressure,o_nitrox_fraction, depth, 6);
      *barliter+=AIR(depth)*(6)/60.0;
      total_time+=6;
      if (o_profile) {
	outs("D");
	outu(total_time);
	outs(" ");
	outu(depth*100);
	outs(" dummy\n");
      }
    }
    if (total_time>ground_time) {
      verbose(ground_time,"ground time shorter than time needed to descend",
	1,total_time/60);
      return 0;
    }
  } else {
    /* pessimistically assume that the diver descends with lightning speed */
    depth=target_depth;
    if (o_profile) {
      outs("D1 ");
      outu(depth*100);
      outs(" dummy\n");
    }
  }
  /* spend ground time */
  zhl16_calc_load(load,o_air_pressure,o_nitrox_fraction, depth, 
    ground_time-total_time);
  *barliter+=AIR(depth)*(ground_time-total_time)/60.0;
  total_time=ground_time;
  if (o_profile) {
    outs("D");
    outu(total_time);
    outs(" ");
    outu(depth*100);
    outs(" dummy\n");
  }

  /* find out whether decompression is needed */

  zhl16_copy(sim_load,load);
  sim_depth=depth;
  while (1) {
    float moved;
    zhl16_calc_load(sim_load,o_air_pressure,o_nitrox_fraction, sim_depth,6);
    moved=o_ascend_speed/60*6;
    sim_depth-=moved;
    if (sim_depth<0) sim_depth=0;

    zhl16_get_max_load_for_depth(maxload, o_air_pressure, o_nitrox_fraction, 
      sim_depth);
    for (g=0;g<gewebe;g++)
      if ((sim_load[g]-baseload)/(maxload[g]-baseload)>o_load_margin)
	break;
    if (g<gewebe) 
      break; /* not safe */
    if (sim_depth<=0) {
      /* no decompression needed */
      verbose(ground_time,"no decompression needed",0,0);
      return 1;
    }
  }

  for (deco_pointer=0;deco_pointer<deco_depths_count;deco_pointer++)
    if (deco_depths[deco_pointer]*feetfactor>sim_depth)
      break;
  if (deco_pointer==deco_depths_count) {
    verbose(ground_time,"deco_depth exceeds scale", 1,sim_depth);
    return 0; /* impossible */
  }
  deco_depth=deco_depths[deco_pointer]*feetfactor;

  if (deco_depth>=depth) {
    verbose(ground_time,"ECANTHAPPEN: deco_depth > depth", 1,deco_depth);
    return 0;
  }

  while (deco_depth>0) {
    unsigned long next_deco_depth;
    /* ascend to deco_depth */
    while (depth>deco_depth) {
      float seconds=(depth-deco_depth)/o_ascend_speed*60;
      float moved;
      if (seconds>6) seconds=6;
      zhl16_calc_load(load,o_air_pressure,o_nitrox_fraction, depth,seconds);
      *barliter+=AIR(depth)*seconds/60.0;
      moved=seconds/60*o_ascend_speed;
      depth-=moved;
      if (depth<deco_depth)
	depth=deco_depth;
      total_time+=seconds;
      if (o_profile) {
	outs("D");
	outu(total_time);
	outs(" ");
	outu(depth*100);
	outs(" dummy\n");
      }
    }
    /* get next deco depth */
    if (0==deco_pointer)
      next_deco_depth=0;
    else
      next_deco_depth=deco_depths[deco_pointer-1]*feetfactor;

    /* stay on deco depth */
    while (1) {
      int ok;
      zhl16_calc_load(load,o_air_pressure,o_nitrox_fraction, depth,60);
      *barliter+=AIR(depth)*60/60.0;
      total_time+=60;
      if (o_profile) {
	outs("D");
	outu(total_time);
	outs(" ");
	outu(depth*100);
	outs(" dummy\n");
      }
      deco_minutes[deco_pointer]++;
      if (o_deco_step_time_limit && 
	  deco_minutes[deco_pointer]>o_deco_step_time_limit) {
	verbose(ground_time,"deco-step-time-limit exceeded",1,
	  deco_minutes[deco_pointer]);
	return 0;
      }

      /* find out whether it's safe to ascend to the next step */
      zhl16_copy(sim_load,load);
      sim_depth=depth;
      ok=0;
      while (1) {
	float moved;

	zhl16_calc_load(sim_load,o_air_pressure,o_nitrox_fraction, sim_depth,6);

	moved=o_ascend_speed/60*6;
	sim_depth-=moved;
	if (sim_depth<next_deco_depth)
	  sim_depth=next_deco_depth;

	zhl16_get_max_load_for_depth(maxload, o_air_pressure, 
	  o_nitrox_fraction, sim_depth);

	for (g=0;g<gewebe;g++)
	  if ((sim_load[g]-baseload)/(maxload[g]-baseload)
	    >=o_load_margin)
	    break;
	if (g<gewebe)
	  break; /* not safe */
	if (fabs(sim_depth-next_deco_depth) <0.1) {
	  ok=1;
	  break;
	}
      }
      if (ok) 
	break;
      continue;
    }
    deco_depth=next_deco_depth;
    deco_pointer--;
  }
  if (o_profile) {
    outs("D");
    outu(total_time+1);
    outs(" 0 dummy\n");
    outs("D");
    outu(total_time*1.1+2);
    outs(" 0 dummy\n");
  }
  return 2;
}

int xx;
char *yy;
#define STRING(x) #x
#define STRING2(x) STRING(x)
static int 
handle_deco_depths(struct uogetopt_env *env, uogetopt2 *x, char *s)
{
  unsigned int i;
  (void) env;
  (void) x;
  deco_depths_count=1;
  for (i=0;s[i];i++)
    if (s[i]==',')
      deco_depths_count++;
  deco_depths=(float*) alloc(deco_depths_count*sizeof(*deco_depths));
  if (!deco_depths) oom();
  deco_depths_count=0;
  for (i=0;s[i];) {
    unsigned int l;
    float f;
    l=scan__float(s+i,&f);
    if (!l || (s[i+l]!=',' && s[i+l]!=0)) 
      xbailout(100,0,"failed to parse ",s," at ",s+i);
    i+=l;
    if (s[i])
      i++;
    if (f<0 || f<0.1)
      xbailout(100,0,"decompression step too low in ",s," at ",s+i);
    deco_depths[deco_depths_count++]=f;
  }
  return 0;
}
static uogetopt2 opts[]={
  {'a',"ascend-speed", do_length_opt,0, &o_ascend_speed,0,
    "Ascend rate in units per minute.",
/*   12345678901234567890123456789012345678901234567890 */
    "Default is " STRING2(DEFAULT_ASCEND_SPEED) "meters. "
    "You have to append a unit to the number, if the "
    "unit isn't meter. A value of 0 is not allowed.\n"
    "The real ascend speed must not be faster than the "
    "value given here. Ascend speeds over 10 meters per "
    "seconds are likely to be unsafe for medical "
    "reasons anyway","LENGTH"},
  {'A',"air-pressure-at-surface", do_pressure_opt,0, &o_air_pressure,1,
    "Air pressure at surface.\n",
    "Append an unit to the number, unless the unit is "
    "the default unit 'bar'. Default value: 1 bar.","PRESSURE"},
  {'d',"descend-speed", do_length_opt,0, &o_descend_speed,0,
    "Descend rate per minute.\n",
    "Default is " STRING2(DEFAULT_DESCEND_SPEED) ". A value of 0 "
    "means that the diver reaches the ground instantly."
    ,"LENGTH"},
  {'D',"desaturation-factor", uogo_float,0, &zhl16_desat_factor,1,
    "Desaturation speed factor.\n",
/*   12345678901234567890123456789012345678901234567890 */
    "Assume that desaturation halfvalue times are "
    "FACTOR times longer as usual. The default FACTOR "
    "is " STRING2(DEFAULT_DECO_FACTOR)". "
    "Values over one create safer tables. Values below "
    "one are highly unsafe, so avoid them.","FACTOR"},
  {  0,"deco-step-time-limit", uogo_ulong,0, &o_deco_step_time_limit,0,
    "Limit decompression times.\n",
    "Stop calculation if the time on any decompression "
    "step exceeds LIMIT minutes. "
/*   12345678901234567890123456789012345678901234567890 */
    "Default: " STRING2(DEFAULT_DECO_STEP_TIME_LIMIT) ".","LIMIT"},
  {  0,"feet", uogo_flag,UOGO_NOARG, &o_feet,1,
    "Use feet, not meters.",
    "For use in backward countries.",0},
  {'L',"load-margin", uogo_float,0, &o_load_margin,0,
/*   12345678901234567890123456789012345678901234567890 */
    "Tissue saturation limit in % of allowed maximum.\n",
    "Try to keep any tissues nitrogen saturation below "
    "NUMBER times 100 percent of the allowed maximum value for the "
    "current depth. The default is " STRING2(DEFAULT_LOAD_MARGIN) ". "
    "The difference between 1 and NUMBER is the safety "
    "margin - your safety margin. Lower values create "
    "safer tables. The default value leads to tables "
    "being a little bit safer than the deco200 table. "
    "\n"
    "The use of higher values is not recommended. Few "
    "depth gauges are perfect, and even less, if any, "
    "divers make no mistakes at all. Even your body "
    "may not work as expected, at least at the day of "
    "your last trip.","NUMBER"},
  {'m',"model", uogo_string,0, &o_zhl16_setting,0,
    "Decompression model to use.\n",
    "The default is " DEFAULT_MODEL ", which was meant to create "
    "dive tables with.\n"
    "Distributed model descriptions include at least 'zhl16a', "
    "'zhl16b', 'zhl16c'.  Model description files are expected to "
    "be named NAME.model, and are searched for in "
    "/package/" PACKAGE_CLASS "/" PACKAGE_VERSION "/data and "
    "the current directory, in that order."
    ,"MODEL"},
  {'M',"make-dive-profile-input", uogo_flag,UOGO_NOARG, 
    &o_mdp,1,"Create input for the make-dive-profile program.\n",
    "Use this for tricks like this:\n"
    "  make-dive-profile `make-deco-table 78 15 15` \\\n"
    "  | mosquitograph -"
    ,0},
  {'n',"nitrox-fraction", uogo_float,0, &o_nitrox_fraction,1,
/*   12345678901234567890123456789012345678901234567890 */
    "The fraction of nitrogen in the cylinder.\n",
    "In other words: Nitrogen partial pressure in gas mix at one bar "
    "ambient pressure. Default: 0.78.",0},
  {  0,"profile", uogo_flag,UOGO_NOARG, &o_profile,1,
    "Generate dive profile.\n",
    "The profile will be usable as input for the mosquitograph program:\n"
    "  make-deco-table 78 15 15 | mosquitograph -\n"
    "This differs from the --make-dive-profile-input by the different "
    "output formats, the data size and the intervals used."
    ,0},
  {'r',"respiratory-minute-volume", do_volume_opt,0, &o_rmv,0,
    "Air usage per minute at sea level (1 bar).\n",
    "make-deco-table will calculate the gas needed for the dive, not "
    "including any reserve, if "
    "this option is given. You may append an unit to the number. "
    "Knows units are liter (default) and cuft.\n"
    "The default is not to calculate gas usage.","VOLUME"},
  {'S',"saturation-factor", uogo_float,0, &zhl16_sat_factor,1,
    "Saturation speed factor.\n",
    "Assume that saturation halfvalue times are FACTOR times as long "
    "as usual. The default FACTOR is 1.\n"
    "Values larger 1 should be avoided at all cost.\n"
/*   12345678901234567890123456789012345678901234567890 */
    "See the --desaturation-factor option for how to\n"
    "slow down desaturation.","FACTOR"},
  {'s',"deco-steps", handle_deco_depths,0, 0,0,
    "Set decompression depths to use.",
    "ARRAY is a list of depths, separated by commas. The program will only "
    "consider these decompression depths and will not try any other. For "
    "example, `--deco-steps 2,4,6,9,12' will let the program try decompression "
    "steps on 2, 4, 6, 9 and 12 meters, but will not allow it to try 15 "
    "meters.\n"
    "When used together with the --feet option the depths in the list will "
    "be treated as feet.\n"
    "The default list contains 3,6,9,12,15,... meters or 10,20,30,40... feet.\n"
    "The values in the list should be in ascending order. Abuse like `--deco-steps 6,3,9' will create annoying and stupid, but safe, decompression rules.\n"
    "This option was added for version 0.1.1 and replaces the --2-meter-steps, --3-meter-steps, --minimal-deco-depth and --maximal-deco-depth options used before.",
    "ARRAY"
  },
  {'V',"verbose", uogo_flag,UOGO_NOARG, &o_verbose,1,
    "Be verbose, show a bit more information.",0,0},
{  0,"examples",uogo_print_help,UOGO_NOARG|UOGO_EXIT|UOGO_NOLHD,
  0,0,
  "Usage examples.",
  "Deco table for 36 meter / 60 minutes, 5 minute intervals:\n"
  "  make-deco-table 36 60 5 5\n"
  "Include gas need for an RMV of 22 liter per minute:\n"
  "  make-deco-table -r 22 36 60 5 5\n"
  "Create a safer table:\n"
  "  make-deco-table --load-margin 0.70 36 60 5 5\n"
  "Calculate decompression rules and feed the output directly into "
  "mosquitograph, this time for a depth of 40 meters and a ground time "
  "of 15 meters:\n"
  "  make-deco-table --profile 40 15 15 | \\\n"
  "  mosquitograph --load-ambient-% - | \\\n"
  "  groffer\n"
  "Calculate decompression rules and feed the output directly into "
  "make-dive-profile, using a depth of 20 meters and 75 minutes ground "
  "time:\n"
  "  make-dive-profile `make-deco-table \\\n"
  "    --make-dive-profile-input 40 75 75 ` | \\\n"
  "    mosquitograph --load-ambient-% - | groffer\n",0},
  {  0,"see-also",uogo_print_help,UOGO_NOARG|UOGO_EXIT|UOGO_NOLHD,
    0,0,
    "Where to find related information.",
    "mosquitograph(1) turns dive profiles into pictures.\n"
    "dive-profile(5) describes the dive profile data format.\n"
    "make-dive-profile(1) creates dive profiles.\n"
    COMMON_RELATED_INFO
    ,0},
  {  0,"caveat",uogo_print_help,UOGO_HIDDEN|UOGO_NOARG|UOGO_EXIT|UOGO_NOLHD,
    0,0,
    "Show Warnings.",
    "Diving near the limits may be safer than skydiving without parachutes, "
    "but that means nothing at all.\n"
    "\n"
    "This software is still in an early testing stage and may contain "
    "serious flaws. Do not bet your life on it: Check it's results "
    "against decompression rules from different sources and against "
    "common sense!\n\n"
    "Note that it's a good idea to include a safety stop even if the "
    "decompression table does not indicate the need for a decompression "
    "stop."
    ,0},
  {  0,"limitations",uogo_print_help,UOGO_NOARG|UOGO_EXIT|UOGO_NOLHD,
    0,0,
    "Show limitations.",
    "make-deco-table is suited for constant gas mixes consisting of oxygen "
    "and nitrogen only.\n\n"
    "make-deco-table only cares about the tissue load with inert gas. "
    "It does not, in any way, deal with the problems resulting from "
    "high oxygen doses.\n\n"
    "make-deco-table does not output decompression stops in feet, but "
    "uses meters. This may change in the future. I feel sorry for "
    "the people in backwards countries.\n\n"
    "make-deco-table has no support at all for more than one dive per day "
    "and doesn't print 'no flying times'.\n\n"
    "make-deco-table is not well-tested at that moment. I'd check tables "
    "generated by it against tables from other sources."
    ,0},
 
  {0,"",uogo_include,0,general_opts, 0,0,0,0},
  {0,0,0,0,0,0,0,0,0}
};

uogetopt_env optenv={
  0,PACKAGE,VERSION,
  "usage: make-deco-table [OPTIONS] DEPTH END-MINUTES [START-MINUTES [STEP]]\n",
  "Create decompression tables\n",
"make-deco-table creates a decompression table suitable for dives of `DEPTH'\n"
"meters and up to `END-MINUTES' minutes of ground time. The program will\n"
"begin calculations with a ground time of `START-MINUTES' (default: 1), but\n"
"will not show ground times not needing any decompression steps.\n"
"`STEP' (default: 1) is the increment to use from `START-MINUTES' to\n"
"`END-MINUTES."
,COMMON_BUGREPORT_INFO,
  3,5,
  0,0,
  uogetopt_out,
  opts
};

static void 
do_header(float depth)
{
  outs("\nDecompression table for ");
  outu(depth);
  outs(o_feet ? " feet" : " meter");
  outs(" depth and a nitrox percentage of ");
  outf(o_nitrox_fraction*100,0);
  outs("%.\n");
  outs("Decompression model used: ");
  outs(o_zhl16_setting);
  outs("\n\n");

  outs("Assumptions:\n");
  if (o_descend_speed>0) {
    outs("* Maximum descend speed: ");
    outf(o_descend_speed/feetfactor,1);
    outs(o_feet ? " feet" : " meter");
    outs("/minute\n");
  } else {
    outs("* Unlimited descend speed.\n");
  }
  outs("* Air pressure about: ");
  outf(o_air_pressure*1000,0);
  outs(" mbar\n");
  outs("* Maximum ascend speed: ");
  outf(o_ascend_speed/feetfactor,1);
  outs(o_feet ? " feet" : " meter");
  outs("/minute\n");
  if (o_rmv>0) {
    outs("* Air usage calculation based on RMV of ");
    outf(o_rmv,1); outs(" liter/minute without any reserve.\n");
  }
  if (fabs(o_load_margin-DEFAULT_LOAD_MARGIN)>0.01) {
    outs("* Tissue load safety margin: ");
    outf(o_load_margin,2);
    if (o_load_margin<DEFAULT_LOAD_MARGIN)
      outs(" (safer)");
    else if (o_load_margin>DEFAULT_LOAD_MARGIN)
      outs(" (less safe)");
    outs("\n");
  }
  if (fabs(zhl16_desat_factor-1)>0.01) {
    outs("* Desaturation factor: ");
    outf(zhl16_desat_factor,3);
    if (zhl16_desat_factor>1)
      outs(" (safer)");
    else if (zhl16_desat_factor<1)
      outs(" (highly unsafe)");
    outs("\n");
  }
  if (fabs(zhl16_sat_factor-1)>0.01) {
    outs("* Saturation factor: ");
    outf(zhl16_sat_factor,3);
    if (zhl16_sat_factor>1)
      outs(" (highly unsafe)");
    else if (zhl16_sat_factor<1)
      outs(" (safer)");
    outs("\n");
  }

  if (o_ascend_speed>10) 
    warn_bad_table("ascend speed too high");
  if (o_load_margin>0.80) 
    warn_bad_table("load margin is dangerously optimistical");
  if (zhl16_desat_factor<1) 
    warn_bad_table("halftime values for decompression to low");
  if (zhl16_sat_factor>1) 
    warn_bad_table("halftime values for compression to high");
  {	
    float x;
    x=depth/10.0;
    if (o_feet) x*=FOOT;
    x=(x+o_air_pressure)*(1-o_nitrox_fraction);
    if (x>1.6)
      warn_bad_table("Oxygen partial pressure dangerously high, over 1.6bar");
    else if (x>1.4)
      warn_bad_table("Oxygen partial pressure to high, over 1.4 bar");
  }
}

int 
main(int argc, char **argv)
{
  unsigned int l;
  unsigned long depth;
  unsigned long ground_time;
  unsigned long t;
  unsigned int gewebe;
  float *load;
  unsigned long *deco_minutes;
  unsigned long start=1;
  unsigned long step=1;
  float barliter;
  int count_no_deco=0;
  int count_deco=0;
  int count_impossible=0;
  unsigned long last_no_deco=0;
  unsigned int max_deco_steps=0;

  uogetopt2 *o;
  bailout_progname(argv[0]);

  flag_bailout_fatal_begin=3;
  bailout_progname(argv[0]);
  optenv.program=flag_bailout_log_name;
  o=uogetopt_join(opts);
  if (!o) oom();
  optenv.opts=o;

  zhl16_desat_factor=DEFAULT_DECO_FACTOR;

  uogetopt_parse (&optenv, &argc, argv);
  zhl16_setting(o_zhl16_setting);
  if (o_ascend_speed<1)
    xbailout(2,0,"--ascend-speed value must be at least 1",0,0,0);

  l=scan_ulong(argv[1],&depth);
  if (!l || argv[1][l])
    xbailout(2,0,"cannot parse depth ",argv[1],0,0);
  l=scan_ulong(argv[2],&ground_time);
  if (!l || argv[2][l])
    xbailout(2,0,"cannot parse ground time ",argv[2],0,0);
  if (argv[3]) {
    l=scan_ulong(argv[3],&start);
    if (!l || argv[3][l])
      xbailout(2,0,"cannot parse start ground time ",argv[3],0,0);
    if (argv[4]) {
      l=scan_ulong(argv[4],&step);
      if (!l || argv[4][l])
	xbailout(2,0,"cannot parse step time ",argv[4],0,0);
    }
  }

  load=zhl16_alloc(&gewebe);
  if (!load) oom();
  baseload=zhl16_init_load(load,o_air_pressure,0.78);

  if (o_feet)
    feetfactor=FOOT;

  if (!deco_depths_count) {
    unsigned int i;
    unsigned int inc=o_feet ? 10 : 3;
    deco_depths_count=depth/feetfactor/inc;
    deco_depths=(float*) alloc(deco_depths_count*sizeof(*deco_depths));
    if (!deco_depths) oom();
    for (i=0;i<deco_depths_count;i++) {
      if (!i) deco_depths[i]=inc;
      else    deco_depths[i]=deco_depths[i-1]+inc;
    }
  }
  deco_minutes=(unsigned long *) alloc(deco_depths_count*sizeof(*deco_minutes));
  if (!deco_minutes) oom();

  if (o_profile) {
    start=ground_time;
    outs("#demo-dive\n");
    outs("*dive0\n");
  }

  if (!o_mdp && !o_profile) {
    for (t=ground_time;t>=1;t-=1) {
      unsigned int i;
      unsigned int res;
      for (i=0;i<deco_depths_count;i++)
	deco_minutes[i]=0;
      res=do_one_dive(depth*feetfactor,t*60,load,deco_minutes,&barliter);
      if (1==res)
	break;
      if (0==res)
	continue;
      do_header(depth);
      outs("\n" "Ground");
      max_deco_steps=0;
      for (i=0;i<deco_depths_count;i++) {
	if (deco_minutes[i]) {
	  outs("  ");
	  outun(deco_depths[i],2);
	  if (o_feet) 
	    outs("f");
	  else
	    outs("m");
	  max_deco_steps++;
	}
      }
      outs(" total\n");
      outs("-time\n");
      break;
    }
  }
  for (t=start;t<=ground_time;t+=step) {
    unsigned int i;
    unsigned int res;
    unsigned long total=0;
    unsigned int this_deco_steps=0;
    for (i=0;i<deco_depths_count;i++)
      deco_minutes[i]=0;
    res=do_one_dive(depth*feetfactor,t*60,load,deco_minutes,&barliter);
    if (!res) {
      count_impossible++;
      continue; /* impossible */
    } else if (1==res)  {
      last_no_deco=t;
      count_no_deco++;
      continue; /* no deco needed */
    }
    count_deco++;

    if (o_mdp) {
      outu(depth);
      outs(" -");
      outu(t);
      i=deco_depths_count-1;
      do {
	if (deco_minutes[i]) {
	  outs("   ");
	  outu(deco_depths[i]);
	  outs(" ");
	  outu(deco_minutes[i]);
	}
      } while (i--);
      outs("\n");
      continue;
    }
    if (o_profile)
      continue;
    outs("  ");
    
    outun(t,3);
    outs(" ");

    for (i=0;i<deco_depths_count;i++) {
      total+=deco_minutes[i];
      l=deco_minutes[i];
      if (l) {
	outs("  ");
	outun(l,3);
	this_deco_steps++;
      }
    }
    while(this_deco_steps++<max_deco_steps)
      outs("     ");
    outs("  ");
    outun(total,4);
    if (o_rmv > 0) {
      outs(" air: ");
      outf(barliter,1); 
    }
    outs("\n");
    outflush();
  }
  if (!count_deco) {
    if (count_no_deco) {
      if (count_impossible) {
	char nb[FMT_ULONG];
	nb[fmt_ulong(nb,last_no_deco)]=0;
	/* that should really only happen in extremly strange cases,
	 * likely to involve --minimal-deco-depth=LARGE.
	 */
	warning(0,"no decompression needed for ground times up to ",
	  nb, " minutes.",0);
	warning(0,"Further ground times needed decompression, but "
	  " no deco rule could be found.",0,0,0);
      } else
	warning(0,"no decompression needed for any ground time given",0,0,0);
    } else if (count_impossible) {
      warning(0,"The ground times given needed decompression, "
	" but no rule was found.",0,0,0);
    }
  }
  outflush();
  return 0;
}
