#include "buffer.h"
#include "bailout.h"
#include "uogetopt.h"
#include "out.h"
#include "scan.h"
#include "divedata.h"
#include "mosquitotools.h"
#include "stdio.h"
#include "scan__float.h"
#include "str.h"
#include <unistd.h>

static unsigned long o_interval=20;
static const char *  o_extra_header;
static const char *  o_id="demo-dive";
static const char *  o_date;
static int           o_aprupt=0;
#define DEFAULT_ASCEND_SPEED 10
#define DEFAULT_DESCEND_SPEED 20
static float o_ascend_speed=DEFAULT_ASCEND_SPEED;
static float o_descend_speed=DEFAULT_DESCEND_SPEED;
#define STRING(x) #x
#define STRING2(x) STRING(x)

static uogetopt2 opts[]={
  {'a',"ascend-speed", do_length_opt,0, &o_ascend_speed,0,
  /* 12345678901234567890123456789012345678901234567890 */
    "Ascend rate in units per minute. Default "STRING2(DEFAULT_ASCEND_SPEED) "m/min.",
    "You have to append a unit to the number, if the "
    "unit isn't meter. A value of 0 is not allowed.",
    "LENGTH-UNIT"},
  {'A',"aprupt", uogo_flag,UOGO_NOARG, &o_aprupt,1,
    "Generate profiles with aprupt depth changes.",
    "The default is to use smooth depth changes.\n"
    "An aprupt profile ignores the --ascend-speed and\n"
    "--descend-speed options.", 0},
  {'d',"descend-speed", do_length_opt,0, &o_descend_speed,0,
    "Descend rate in units per minute. Default "STRING2(DEFAULT_DESCEND_SPEED) "m/min.",
    "You have to append a unit to the number, if the "
    "unit isn't meter. A value of 0 is not allowed.",
    "LENGTH-UNIT"},
  {'D',"date", uogo_string, 0, &o_date,0,
    "Use date as dive-start-date.",
    "The date and time have to be given in the iso-8601 format. This option "
    "is useful when you want to use the continuous load calculation option "
    "of mosquitograph.\n"
    "The date format is something like '2003-07-20T13:14'.",
    "YYYY-MM-DDThh:mm[:ss]"},
  {'I',"id", uogo_string, 0, &o_id,0,
    "Use ID as dive identifier in the profile.",
    "The default is 'demo-dive', which should usually be good enough. "
    "You may find it useful to use different dive ids if you plan to "
    "work with more than one dive profile created by make-dive-profile."
    ,"ID"},
  {'i',"interval", uogo_ulong,0, &o_interval,0,
    "Fake sampling interval.",
    "This interval determines the time difference between two values in "
    "the dive profile and the granularity of the graph. The default "
    "interval is 20 seconds. When given an interval of 0 make-dive-profile "
    "will create a profile with just the absolutely needed points, which "
    "will be much shorter, but not as beautiful.", "SECONDS"},
  {  0,"see-also",uogo_print_help,UOGO_NOARG|UOGO_EXIT|UOGO_NOLHD,
    0,0,
    "Where to find related information.",
    "mosquitograph(1) turns dive profiles into pictures.\n"
    "dive-profile(5) describes the dive profile data format.\n"
    "mosquitoget(1) reads dive profiles from suunto mosquito dive computers.\n"
    COMMON_RELATED_INFO
    ,0},
  {  0,"examples",uogo_print_help,UOGO_NOARG|UOGO_EXIT|UOGO_NOLHD,
    0,0,
    "Usage examples.",
    "Create a dive profile with a ground time of 15 minutes on 40 meters, followed by 5 minutes on 12, 5 minutes on 9, 5 minutes on 6 and 10 minutes on 3 meters:\n"
    "  make-dive-profile 40 -15   12 5   9 5   6 5   3 10\n"
    "Now do the same using a slow ascend speed of 2 meters per minute:\n"
    "  make-dive-profile -a 2 40 -15   12 5   9 5   6 5   3 10\n"
    ,0},

  {0,"",uogo_include,0,general_opts, 0,0,0,0},
  {0,0,0,0,0,0,0,0,0}
  };
static uogetopt_env optenv={
  0,PACKAGE,VERSION,
    "usage: make-dive-profile depth time [depth time pairs...]\n",
"Create a dive profile.",
"make-dive-profile creates a dive profile from the depth / time pairs on the\n"
"command line for later use with mosquitograph.\n"
"It may be useful for education and to answer 'what if' questions.\n\n"
"'depth' is the depth of each step in meters, and 'time' is it's length.\n"
"A negative 'time' means to stay on 'depth' until 'time' minutes have\n"
"passed since the start of the dive.\n\n"
"A unit may be appended to 'depth' (defaults to meter, use the --show-units\n"
"option for more units) and 'time' (defaults to minutes, but 'seconds' and\n"
"hours are also recognized).\n\n"
"To switch gas during the `dive' use 'n2 XX' instead of a depth/time pair, "
"where N2 is the new nitrogen percentage of the gas mix. The default gas "
"is air, containing 78% nitrogen. To use a different gas from the beginning "
"of the dive use 'n2 ZZ' as the first depth/time pair.",
COMMON_BUGREPORT_INFO,
    2,0,
    0,0,
    uogetopt_out,
    opts
};
static short n2=101;

static void line(unsigned long time, unsigned long depth)
{
  outs("D");
  outu(time);
  outs(" ");
  outu(depth);
  outs(" dummy");
  if (n2!=101) {
    outs(" EAN");
    outu(100-n2);
    n2=101;
  }
  outs("\n");
}
static unsigned long 
get_depth(char *s)
{
  float x;
  uogetopt2 o={0,0,do_length_opt,0,&x,0,0,0,0};
  if (do_length_opt(&optenv,&o,s))
    _exit(1);
  return x*100;
}

int main(int argc, char **argv)
{
  unsigned int i;
  unsigned long now;
  unsigned long last_depth;
  uogetopt2 *o;

  flag_bailout_fatal_begin=3;
  uogo_posixmode=1;

  bailout_progname(argv[0]);
  o=uogetopt_join(opts);;
  if (!o) oom();
  optenv.program=flag_bailout_log_name;
  optenv.opts=o;
  uogetopt_parse (&optenv, &argc, argv);
  if ((1+argc)%2)
    xbailout(100,0,"need 'depth time' pairs.",0,0,0);
  outs("#"); 
  outs(o_id);
  outs("\n");
  outs("*dive0\n");
  if (o_date) {
    outs(":dive-start-date ");
    outs(o_date);
    outs("\n");
  }
  if (o_extra_header)
    outs(o_extra_header);
  now=0;
  last_depth=0;
  for (i=1;;i+=2) {
    unsigned long de;
    float ti;
    unsigned int l;
    const char *s;
    int minus=0;
    if (argv[i] && str_equal(argv[i],"n2")) {
      unsigned long ul;
      l=scan_ulong(argv[i+1],&ul);
      if (!l || argv[i+1][l] || ul>100)
	xbailout(100,0,"failed to parse n2 percentage ",argv[i+1],0,0);
      n2=ul;
      continue;
    }
    if (now==0)
      line(0,0);
    if (argv[i]) {
      de=get_depth(argv[i]);

      s=argv[i+1];
      if (*s=='-') {
	s++;
	minus=1;
      }
      l=scan__float(s,&ti);
      if (!l) 
	xbailout(100,0,"cannot parse time ",argv[i+1],0,0);
      if (s[l]=='s')
	ti=ti;
      else if (s[l]=='m'|| !s[l])
	ti=ti*60;
      else if (s[l]=='h')
	ti=ti*3600;
      else
	xbailout(100,0,"cannot parse time ",argv[i+1],0,0);
    } else if (last_depth==0) {
      break;
    } else {
      de=0;
      ti=60;
    }

    if (!o_aprupt && de !=last_depth) {
      unsigned long in=o_interval;
      unsigned int d=last_depth;
      if (!in) in=20;
      if (de>d) {
	while (de>d) {
	  unsigned long t_old;
	  float delta=o_descend_speed*100.0/60*in;
	  if (d+delta<de) {
	    d+=delta;
	    now+=in;
	    line(now,d);
	    continue;
	  }
	  t_old=now;
	  now+=(de-d)/(o_descend_speed*100/60);
	  if (now==t_old)
	    now++;
	  line(now,de);
	  break;
	}
	d=de;
      } else {
	while (de<d) {
	  unsigned long t_old;
	  float delta=o_ascend_speed*100.0/60*in;
	  if (d-delta>de) {
	    d-=delta;
	    now+=in;
	    line(now,d);
	    continue;
	  }
	  t_old=now;
	  now+=(d-de)/(o_ascend_speed*100/60);
	  if (now==t_old)
	    now++;
	  line(now,de);
	  break;
	}
	d=de;
      }
    }

    if (minus) {
      if (ti>now)
	ti-=now;
      else
	ti=20;
    }
    if (o_interval) {
      ti+=now;
      while (now<ti) {
	if (now+o_interval<=ti)
	  now+=o_interval;
	else
	  now=ti;
	line(now,de);
      }
    } else {
      now+=ti;
      line(now,de);
    }
    last_depth=de;
    if (!argv[i])
      break;
  }
  outflush();
  return 0;
}
