#include "bailout.h"
#include "buffer.h"
#include "error.h"
#include "str.h"
#include "open.h"
#include "fmt.h"
#include "scan.h"
#include "uogetopt.h"
#include "ndelay.h"
#include "dtr.h"
#include "rts.h"
#include "msleep.h"
#include "iopause.h"
#include "attributes.h"
#include "stralloc.h"
#include "out.h"
#include "mosquitotools.h"
#include <termios.h>
#include <unistd.h>

#define CONF_POS 0
#define CONF_LEN 113

#define PROFILE_WAIT0 15 /* seconds to wait for first profile */
#define PROFILE_WAIT   5

enum model {
  M_SPYDER, M_VYPER_OR_COBRA, M_STINGER, M_VYPER, M_MOSQUITO
};
enum model model;

static int o_debug=0;
static int o_verbose=0;
static int eat_echo;

static int o_profiles=0;

static int o_all=0;
static int o_to_stdout=0;
static int o_conf=0;
static const char *o_dir=".";
static const char *o_buddies;
static const char *o_comment;
static const char *o_equipment;
static const char *o_place;
static float o_bottle_volume=0;
static float o_pressure_at_begin=0;
static float o_pressure_at_end=0;
#ifdef __OpenBSD__
#define DEFAULT_DEVICE  "/dev/tty00"
#else
#define DEFAULT_DEVICE  "/dev/ttyS0"
#endif	
static const char *dev_name=DEFAULT_DEVICE;
static int         dev_fd=-1;
static struct termios tios_orig;
static struct termios tios_now;
static void 
die_restore(int ex, int er, const char *s1, const char *s2,
  const char *s3, const char *s4) attribute_noreturn;

static void 
die_restore(int ex, int er, const char *s1, const char *s2,
  const char *s3, const char *s4)
{
  if (-1==tcsetattr(dev_fd, TCSANOW, &tios_orig))
    warning(errno,"failed to restore terminal attributes",0,0,0);
  xbailout(ex,er,s1,s2,s3,s4);
}
#define F_ONOFF(flag,what) \
static void my_##flag##_##what(void) { \
  if (-1==flag##_##what(dev_fd)) \
    die_restore(111,errno,"failed to turn ",#flag," ",#what); }
F_ONOFF(DTR,on) F_ONOFF(DTR,off) F_ONOFF(RTS,on) F_ONOFF(RTS,off)
static void oom_restore(void) attribute_noreturn;
static void oom_restore(void)
{ die_restore(111,0,"out of memory",0,0,0); }

static void 
dev_config(void)
{
  tcflush(dev_fd, TCIOFLUSH);
  if (-1==tcgetattr(dev_fd, &tios_orig))
    xbailout(111,errno,"failed to get terminal attributes",0,0,0);
  tios_now=tios_orig;
  /* 2400 8 odd 1 stop */
  tios_now.c_cflag=CS8 | CREAD | CLOCAL | PARODD | PARENB;
  tios_now.c_iflag=IGNBRK;
  tios_now.c_lflag=0;
  tios_now.c_oflag=0;
  tios_now.c_cc[VMIN]=0;
  tios_now.c_cc[VTIME]=0;
  cfsetspeed(&tios_now,B2400);
  if (-1==tcsetattr(dev_fd, TCSANOW, &tios_now))
    die_restore(111,errno,"failed to set terminal attributes",0,0,0);
  my_DTR_on();
  msleep(100); /* sleep 100 ms */
  if (o_verbose) warning(0,"device ",dev_name," configured, +DTR",0);
}

static void
dev_open(void)
{
  dev_fd=open_readwrite(dev_name);
  if (-1==dev_fd)
    xbailout(111,errno,"failed to open ",dev_name,0,0);
  if (-1==ndelay_off(dev_fd))
    xbailout(111,errno,"failed to change ",dev_name," to blocking mode",0);
  if (o_verbose) warning(0,"device ",dev_name," opened",0);
  dev_config();
}

static void 
dev_write(const char *s, unsigned int len)
{
  iopause_fd p;
  struct taia now,then;

  taia_now(&now);

  taia_now(&now);
  taia_uint(&then,1);
  taia_half(&then,&then); /* 0.5 */
  taia_add(&then,&now,&then);
  p.fd=dev_fd;
  p.events=IOPAUSE_WRITE;

  if (o_verbose) warning(0,"  +RTS",0,0,0);
  my_RTS_on();
  msleep(500); /* XXX */
  while (len) {
    iopause(&p,1,&then,&now);
    if (p.revents & IOPAUSE_WRITE) {
      int x;
      x=write(dev_fd,s,len);
      if (x<0) {
	if (errno==error_intr)
	  continue;
	die_restore(111,errno,"failed to write to device",0,0,0);
      }
      len-=x;
      s+=len;
      continue;
    }
    taia_now(&now);
    if (taia_less(&then,&now))
      die_restore(111,errno,"timed out writing to device",0,0,0);
  }
  tcdrain(dev_fd);
  msleep(200); /* XXX */
  if (o_verbose) warning(0,"  -RTS",0,0,0);
  my_RTS_off();
}

static int 
wait_for_input(unsigned int msec)
{
  struct taia now,then;
  iopause_fd p;

  taia_now(&now);
  taia_uint(&then,msec/1000);
  then.nano=(msec%1000)*1000000;
  then.atto=0;
  taia_add(&then,&now,&then);
  p.fd=dev_fd;
  p.events=IOPAUSE_READ;
  iopause(&p,1,&then,&now);
  if (o_verbose) 
    warning(0,"wait for input returns ",
      (p.revents & IOPAUSE_READ) ? "input" : "timeout",0,0);
  if (p.revents & IOPAUSE_READ)
    return 1;
  return 0;
}

static int
dev_readchar(unsigned char *c)
{
  iopause_fd p;
  struct taia now,then;

  taia_now(&now);
  taia_uint(&then,1);
  taia_add(&then,&now,&then);
  p.fd=dev_fd;
  p.events=IOPAUSE_READ;

  while (1) {
    iopause(&p,1,&then,&now);
    if (p.revents & IOPAUSE_READ) {
      int x;
      x=read(dev_fd,c,1);
      if (x<0) {
	if (errno==error_intr)
	  continue;
	die_restore(111,errno,"failed to read from device",0,0,0);
      }
      if (!x)
	return 0;
      return 1;
    }
    taia_now(&now);
    if (taia_less(&then,&now))	
      return -1;
  }
}
static void die_eof_restore(void) attribute_noreturn;
static void die_eof_restore(void) 
{ die_restore(111,0,"unexpected end of file reading from device",0,0,0); }
static void die_noecho_restore(unsigned char, unsigned char) attribute_noreturn;
static void die_noecho_restore(unsigned char x, unsigned char g) 
{ 	
  char nb1[FMT_ULONG];
  char nb2[FMT_ULONG];
  nb1[fmt_ulong(nb1,x)]=0;
  nb2[fmt_ulong(nb2,g)]=0;
  die_restore(111,0,"failed to read answer/echo: expected #",nb1,
    ", but got #",nb2);
}
static void
x_dev_readchar(unsigned char *c)
{
  int r=dev_readchar(c);
  if (-1==r) 	
    die_restore(111,0,"timed out",0,0,0);
  if (!r) 
    die_eof_restore();
}
static void 
expect(const char *p, unsigned int len)
{
  if (o_verbose)
    warning(0,"expecting answer",0,0,0);
  while (len) {
    char c;
    x_dev_readchar(&c);
    if (c!=*p)
      die_noecho_restore(*p,c);
    p++;
    len--;
  }
  if (o_verbose)
    warning(0,"expect: got it",0,0,0);
}
static int 
expect_or_not(const char *p, unsigned int len)
{
  unsigned int count=0;
  if (o_verbose)
    warning(0,"expect-or-not start",0,0,0);
  while (len) {
    char c;
    int r;
    r=dev_readchar(&c);
    if (!r)
      die_eof_restore();
    if (-1==r)
      if (count==0) {
	if (o_verbose) warning(0,"expect-or-not: timed out -> no echo",0,0,0);
	return 0;
      }
    if (-1==r)
      die_restore(111,0,"timed out",0,0,0);
    count++;
    if (c!=*p)
      die_noecho_restore(*p,c);
    p++;
    len--;
  }
  if (o_verbose) warning(0,"expect-or-not: got it",0,0,0);
  return 1;
}
static void 
dev_test(void)
{
#if 0
  if (o_verbose) warning(0,"testing device ",dev_name,0,0);
  write(dev_fd,"AT\r",3);
  eat_echo=expect_or_not("AT\r",3);
  if (o_verbose) warning(0,"test of device ",dev_name,": success",0);
#endif

  if (o_verbose) warning(0,"testing device ",dev_name,0,0);
  dev_write("AT\r",3);
  eat_echo=expect_or_not("AT\r",3);
  if (o_verbose) 
    warning(0,"test of device ",dev_name,": success. Interface ",
    eat_echo ? "with echo" : "without echo");
}
static stralloc o_mem;
static int 
callback_mem(uogetopt_env *e, uogetopt2 *g, char *s)
{
  unsigned int x;
  unsigned long p;
  unsigned long l;
  (void) g;
  (void) e;
  x=scan_ulong(s,&p);
  if (s[x]!=':')
    goto bad;
  s+=x+1;
  x=scan_ulong(s,&l);
  if (s[x])
    goto bad;
  if (!l) goto bad;
  if (!stralloc_catb(&o_mem,(void *)&p,sizeof(unsigned long))) oom();
  if (!stralloc_catb(&o_mem,(void *)&l,sizeof(unsigned long))) oom();
  return 0;
bad:
  xbailout(100,0,"bad use of --get-mem, should be '-m offset:len'",0,0,0);
}
static unsigned int 
checksum(unsigned long startval, const char *p, unsigned int l)
{
  unsigned int i;
  unsigned int cs=startval;
  for (i=0;i<l;i++)
    cs^=p[i];
  return cs;
}
static void crc_check(unsigned char icrc, unsigned char gcrc)
{
  if (icrc!=gcrc) {
    char nb1[FMT_ULONG];
    char nb2[FMT_ULONG];
    nb1[fmt_ulong(nb1,icrc)]=0;
    nb2[fmt_ulong(nb2,gcrc)]=0;
    die_restore(111,0,"checksum mismatch. Expected #",nb1,
      ", but got #",nb2);
  }
}

/* the real low level mem get */
static void
mem_block_get(unsigned long pos, unsigned int len, char *p)
{
  unsigned char gcrc;
  unsigned char icrc;
  char cmd[5];
  if (pos > 65536)
    xbailout(100,0,"internal error: pos out of rangen in mem_block_get",0,0,0);
  if (len > 32)
    xbailout(100,0,"internal error: len out of rangen in mem_block_get",0,0,0);
  cmd[0]=5; /* read */
  cmd[1]=pos/256;
  cmd[2]=pos%256;
  cmd[3]=len;
  cmd[4]=checksum(0,cmd,4);
  if (o_debug) {
    char nb1[FMT_ULONG];
    char nb2[FMT_ULONG];
    nb1[fmt_ulong(nb1,pos)]=0;
    nb2[fmt_ulong(nb2,len)]=0;
    warning(0,"debug: mem-block-get from pos ",nb1, " len ",nb2);
  }
  dev_write(cmd,5);
  if (eat_echo) expect(cmd,5); /* echo */
  wait_for_input(PROFILE_WAIT0*1000); /* extra wait */
  if (eat_echo) expect(cmd,4); /* data block start */
  icrc=cmd[4];
  while (len) {
    x_dev_readchar(p);
    icrc=checksum(icrc,p,1);
    p++;
    len--;
  }
  x_dev_readchar(&gcrc);
  crc_check(icrc,gcrc);
  if (o_debug)
    warning(0,"debug: mem-block-get ok",0,0,0);
}

static void 
do_one_mem_range(unsigned long pos, unsigned long len, char *p)
{
  while (len) {
    unsigned int l;
    l=len > 32 ? 32 : len;
    mem_block_get(pos,l,p);
    len-=l;
    pos+=l;
    p+=l;
  }
}
static buffer ob;
static char obspace[BUFFER_OUTSIZE];
static buffer nullb=BUFFER_INIT((buffer_op_t)read,-1,obspace,sizeof(obspace));

static void
do_mem_print(unsigned long pos, unsigned long len, unsigned char *p)
{
  unsigned long i;
  outs("% off 00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f\n");
  for (i=0;i<len;i++) {
    if (i%16==0) {
      if (i)
	outs("\n");
      outs("R");
      outx(i+pos,4);
      outs(" ");
    }
    outx(p[i],2);
    outs(" ");
  }
  if (i)
    outs("\n");
}

static stralloc config_memory;

static void
get_config_memory(void)
{
  if (!stralloc_ready(&config_memory,CONF_LEN)) oom_restore();
  do_one_mem_range(CONF_POS,CONF_LEN,config_memory.s);
  config_memory.len=CONF_LEN;

  switch((unsigned char) config_memory.s[0x24]) {
  case 20:
  case 30:
  case 60: model=M_SPYDER; break;
  case 3: model=M_STINGER; break;
  case 4: model=M_MOSQUITO; break;
  case 10: model=M_VYPER; break;
  case 12: model=M_VYPER_OR_COBRA; break;
  default:
      die_restore(100,0,"failed to determine computer type",0,0,0);
  }
  if (model==M_SPYDER)
    die_restore(100,0,"Spyder handling not implemented",0,0,0);
}

static void
do_mem(void)
{
  unsigned int ranges=o_mem.len/(2*sizeof(unsigned long));
  unsigned int i;
  for (i=0;i<ranges;i++) {
    unsigned long rpos;
    unsigned long rlen;
    stralloc sa=STRALLOC_INIT;
    rpos=*(unsigned long *)(o_mem.s+(2*i*sizeof(unsigned long)));
    rlen=*(unsigned long *)(o_mem.s+((2*i+1)*sizeof(unsigned long)));
    if (!stralloc_ready(&sa,rlen)) oom_restore();
    out_set_buffer(buffer_1);
    do_one_mem_range(rpos,rlen,sa.s);
    outs("#mem-");
    outu(rpos);
    outs("-");
    outu(rlen);
    outs("\n");
    outs("*mem\n");
    do_mem_print(rpos,rlen,sa.s);
    stralloc_free(&sa);
    outflush();
    out_set_buffer(&nullb);
  }
}

static void
do_conf_sub(void)
{
  const unsigned char *s;
  int mode=0;
  unsigned long ul;
  float f;
  unsigned int i;
  enum model m=0; /* keep gcc quiet */

  outs("#conf\n*config0\n");

  do_mem_print(0,config_memory.len,config_memory.s);
  s=config_memory.s;
  switch(s[0x24]) {
  case 20:
  case 30:
  case 60:
    outs(":conf-type Spyder\n");
    mode=1;
    break;
  case 3: outs(":conf-type Stinger\n"); m=M_STINGER; break;
  case 4: outs(":conf-type Mosquito\n"); m=M_MOSQUITO; break;
  case 10: outs(":conf-type Vyper (new)\n"); m=M_VYPER; break;
  case 12: outs(":conf-type Vyper/Cobra\n"); m=M_VYPER_OR_COBRA; break;
  default: outs(":conf-type unknown\n"); mode=-1; break;
  }
  if (-1==mode)
    return;
  if (1==mode)
    return;
  f=s[0x1e]*256+s[0x1f];
  f/=128;
  f*=FOOT;
  outs(":conf-max-depth "); outf(f,2); outs("\n");
  ul=s[0x20]*256+s[0x21];
  outs(":conf-total-dive-time "); outu(ul); outs("\n");
  ul=s[0x22]*256+s[0x23];
  outs(":conf-total-dives "); outu(ul); outs("\n");
  outs(":conf-user "); 
  for (i=0x2c;i<0x4a;i++) {
    if (!s[i]) break;
    outb(s+i,1);
  }
  outs("\n");
  outs(":conf-recording-interval "); outu(s[0x53]); outs("\n");
  outs(":conf-altitude-setting "); outu(s[0x54]%3); outs("\n");
  outs(":conf-personal-setting "); outu(s[0x54]/3); outs("\n");
  f=s[0x5c]*256+s[0x5d];
  f/=128.0;
  f*=FOOT;
  outs(":conf-max-freediving-depth "); outf(f,2); outs("\n");
  ul=s[0x5e]*256+s[0x5f];
  outs(":conf-max-freediving-time "); outu(ul); outs("\n");
  switch(s[0x62]) {
  case 0: outs(":conf-unit-mode imperial\n"); break;
  case 1: outs(":conf-unit-mode metric\n"); break;
  default:outs(":conf-unit-mode unknown\n"); break;
  }
  outs(":conf-operating-mode "); outx(s[0x63],2);
  if (m==M_MOSQUITO || m==M_STINGER) {
    if (s[0x63] & 128) outs(" free");
    if (s[0x63] & 8) outs(" diving");
  } else {
    switch(s[0x63]) {
    case 0: outs(" Air"); break;
    case 1: outs(" Nitrox/EAN"); break;
    case 2: outs(" Gauge"); break;
    default: outs(" unknown");
    }
  }
  outs("\n");
  outs(":conf-light-mode "); outx(s[0x64],2);
  if (s[0x64]&128) outs(" on "); else outs(" off ");
  outu(s[0x64] & 0x7f);
  outs(" seconds\n");
  outs(":conf-alarm-mode "); outx(s[0x65],2);
  if (s[0x65] & 1) outs(" dive-time");
  if (s[0x65] & 2) outs(" dive-depth");
  if (s[0x65] & (~3)) outs(" ?");
  outs("\n");

  ul=s[0x66]*256+s[0x67];
  outs(":conf-dive-alarm-time "); outu(ul); outs("\n");
  f=s[0x68]*256+s[0x69];
  f/=128;
  f*=FOOT;
  outs(":conf-dive-alarm-depth "); outf(f,2); outs("\n");
}

static void
do_conf(void)
{
  out_set_buffer(buffer_1);
  do_conf_sub();
  outflush();
  out_set_buffer(&nullb);
}

static int
print_one_profile(stralloc *sa)
{
  unsigned int i;
  unsigned int l=sa->len;
  unsigned int depth=0;
  unsigned int tim;
  unsigned int interval;
  const unsigned char *s=sa->s;
  unsigned int year;
  static stralloc fn;
  static stralloc fnt;
  static stralloc date;
  unsigned int maxdepth=0;
  int fd;
  int is_freedive=(s[l-5] & 64)!=0;
  year=s[l-10] >90 ? 1990+s[l-10] : 2000+s[l-10];

  if (!stralloc_copys(&date,"")) oom_restore();
  if (!stralloc_catuint0(&date,year,4)) oom_restore();
  if (!stralloc_cats(&date,"-")) oom_restore();
  if (!stralloc_catuint0(&date,s[l-11],2)) oom_restore();
  if (!stralloc_cats(&date,"-")) oom_restore();
  if (!stralloc_catuint0(&date,s[l-12],2)) oom_restore();
  if (!stralloc_cats(&date,"T")) oom_restore();
  if (!stralloc_catuint0(&date,s[l-13],2)) oom_restore();
  if (!stralloc_cats(&date,":")) oom_restore();
  if (!stralloc_catuint0(&date,s[l-14],2)) oom_restore();
  
  if (o_to_stdout)
    fd=1;
  else {
    if (o_dir && *o_dir) {
      if (!stralloc_copys(&fn,o_dir)) oom_restore();
      if (!stralloc_cats(&fn,"/")) oom_restore();
    }  else
      if (!stralloc_copys(&fn,"")) oom_restore();
    if (!stralloc_cat(&fn,&date)) oom_restore();
    if (!stralloc_cats(&fn,"-")) oom_restore();
    if (!stralloc_catuint0(&fn,s[l-3],0)) oom_restore();
    if (!stralloc_cats(&fn,is_freedive ? ".freedive" : ".dive")) oom_restore();
    if (!stralloc_copy(&fnt,&fn)) oom_restore();
    if (!stralloc_0(&fn)) oom_restore();
    if (!stralloc_cats(&fnt,".tmp")) oom_restore();
    if (!stralloc_0(&fnt)) oom_restore();
    fd=open_read(fn.s);
    if (-1!=fd) {
      close(fd);
      warning(0,fn.s," already exists, not overwritten",0,0);
      return 0;
    }
    if (errno!=error_noent)
      /* whatever */
      die_restore(111,errno,"failed to open ",fn.s,0,0);
    fd=open_trunc(fnt.s);
    if (-1==fd)
      die_restore(111,errno,"failed to create/truncate ",fnt.s,0,0);
  }
  buffer_init(&ob,(buffer_op_t) write,fd,obspace,sizeof(obspace));
  out_set_buffer(&ob);
  do_conf_sub();
  outs("#");
  outb(date.s,date.len);
  outs(is_freedive ? "-free" : "-dive");
  outs("\n");
  outs("*dive0\n");
  for (i=0;i<l;i++) {
    if (i%16==0) {
      if (i) outs("\n");
      outs("R");
    }
    if (i%16)
      outs(" ");
    outx(s[i],2);
  }
  if (i)
    outs("\n");
  outs(":dive-mode "); 
  if (is_freedive) outs("free\n");
  else if (s[l-7]) outs("nitrox\n");
  else outs("air\n");

  outs(":surface-interval-minutes "); outu(s[l-1]); outs("\n");
  outs(":surface-interval-hours "); outu(s[l-2]); outs("\n");
  outs(":dive-no "); outu(s[l-3]); outs("\n");
  outs(":recording-interval "); outu(s[l-4]); outs("\n");
  interval=s[l-4];
  outs(":altitude-setting "); outu(s[l-5]%3); outs("\n");
  outs(":personal-setting "); outu(s[l-5]/3); outs("\n");
  if (s[l-6]) {
    outs(":pressure-at-begin "); outu(s[l-6]); outs("\n");
  }
  if (s[1]) {
    outs(":pressure-at-end "); outu(s[1]); outs("\n");
  }
  outs(":oxygen-percentage "); outu(s[l-7]); outs("\n");
  outs(":OLF-value "); outu(s[0]); outs("\n");
  outs(":unknown-dive-value-07 "); outu(s[l-8]); outs("\n");
  outs(":temperature-at-begin "); outu(s[l-9]); outs("\n");
  outs(":temperature-at-max-depth "); outu(s[3]); outs("\n");
  outs(":temperature-at-end "); outu(s[2]); outs("\n");
  outs(":dive-start-date ");
    outu(s[l-10] >90 ? 1990+s[l-10] : 2000+s[l-10]);
    outs("-");
    outu0(s[l-11],2);
    outs("-");
    outu0(s[l-12],2);
    outs("T");
    outu0(s[l-13],2);
    outs(":");
    outu0(s[l-14],2);
    outs("\n");
  depth=0;
  tim=0;
#define MAY_NEED_DUMMY() if (!tim) outs("D0 0 0");
  for (i=l-15;i>3;i--) {
    unsigned char c=s[i];
    if (c>=0x78 && c<=0x87) {
      if      (c==0x7a) { MAY_NEED_DUMMY(); outs(" SLOW"); }
      else if (c==0x7b) { MAY_NEED_DUMMY(); outs(" ATTENTION"); }
      else if (c==0x7c) { MAY_NEED_DUMMY(); outs(" BOOKMARK"); }
      else if (c==0x7d) { MAY_NEED_DUMMY(); outs(" SURFACED"); }
      else if (c==0x7e) { MAY_NEED_DUMMY(); outs(" DECO-STOP-NEEDED"); }
      else if (c==0x7f) { MAY_NEED_DUMMY(); outs(" DECO-STOP-VIOLATION"); }
      else if (c==0x80) break;
      else if (c==0x81) { MAY_NEED_DUMMY(); outs(" SAFETY-STOP-VIOLATION"); }
      else if (c==0x87 && i>3) { 
	MAY_NEED_DUMMY(); 
	outs(" EAN");
	outu(s[++i]);
      }
      else              { MAY_NEED_DUMMY(); outs(" UNKNOWNx"); outx(c,2); }
	continue;
    } else if (c<0x78)
      depth+=c;
    else
      depth-=256-c;

    if (i!=l-15) outs("\n");
    outs("D");
    tim+=interval;
    outu(tim);
    outs(" ");
    outu( ((float) depth)*FOOT *100);
    outs(" "); outx(depth,2);
    if (depth>maxdepth) maxdepth=depth;
  }
  if (i)
    outs("\n");
  outs(":max-depth "); outf(maxdepth*FOOT,2); outs("\n");
  outs(":total-seconds "); outu(tim); outs("\n");

  outs("%\n% you may want to edit these:\n%\n");

  outs("% your dive number\n");
  outs(":total-dive-number \n");

  outs("% bottle volume in liter.\n");
  outs(":bottle-volume ");
  if (o_bottle_volume>0)
    outu(o_bottle_volume);
  outs("\n");

  if (!s[l-6]) {
    outs("% pressure at start of dive.\n");
    outs(":pressure-at-begin ");
    if (o_pressure_at_begin>0)
      outu(o_pressure_at_begin);
    outs("\n");
  } else if (o_pressure_at_begin>0)
    warning(0,"start-pressure option ignored",0,0,0);

  if (!s[1]) {
    outs("% pressure at end of dive.\n");
    outs(":pressure-at-end ");
    if (o_pressure_at_end>0)
      outu(o_pressure_at_end);
    outs("\n");
  } else if (o_pressure_at_end>0)
    warning(0,"end-pressure option ignored",0,0,0);

  outs(":place ");
  if (o_place)
    outs(o_place);
  outs("\n");
  outs(":equipment ");
  if (o_equipment)
    outs(o_equipment);
  outs("\n");
  outs(":buddies ");
  if (o_buddies)
    outs(o_buddies);
  outs("\n");
  outs(":comment0 ");
  if (o_comment)
    outs(o_comment);
  outs("\n");
  outs("%:comment1 \n");
  outs("% and so on, but numbers must be continuous\n");

  outflush();
  out_set_buffer(&nullb);
  if (1!=fd) {
    if (-1==fsync(fd)) 
      die_restore(111,errno,"failed to fsync ",fnt.s,0,0);
    if (-1==close(fd)) 
      die_restore(111,errno,"failed to close ",fnt.s,0,0);
    if (-1==rename(fnt.s,fn.s)) 
      die_restore(111,errno,"failed to rename ",fnt.s," to ",fn.s);
    warning(0,"got profile ",fn.s,0,0);
  }
  return 1;
}

static int
one_profile(int first)
{
  char cmd[3];
  unsigned char count;
  static stralloc sa;
  unsigned int i;

  cmd[0]=first ? 8 : 9;
  cmd[1]=165;
  cmd[2]=checksum(0,cmd,2);
  if (o_debug)
    warning(0,"debug: one_profile ",first? "first" : "next" ,0,0);
  sa.len=0;

  if (o_verbose) warning(0,"trying to get profile",0,0,0);
  dev_write(cmd,3);
  if (eat_echo) expect(cmd,3); /* echo */
  if (o_verbose) warning(0,"extra wait",0,0,0);
  wait_for_input(PROFILE_WAIT*1000); /* extra wait */
  while (1) {
    unsigned char gcrc,icrc;
    if (eat_echo) expect(cmd,1); /* data block start */
    icrc=checksum(0,cmd,1);
    x_dev_readchar(&count);
    if (count>32)
      die_restore(111,0,"bad profile data block size",0,0,0);
    icrc=checksum(icrc,&count,1);
    for (i=0;i<count;i++) {
      unsigned char c;
      x_dev_readchar(&c);
      if (!stralloc_append(&sa,&c)) oom_restore();
      icrc=checksum(icrc,&c,1);
    }
    x_dev_readchar(&gcrc);
    crc_check(icrc,gcrc);
    /* oh, yeah - to see whether the profile continues i use a wait?
     * shouldn't i scan the profile for the end of profile marker? 
     */
    if (0==count && !sa.len)
      return 0; /* null profile - end of profiles */
    if (!wait_for_input(500))
      break;
  }
  if (o_verbose) {
    char nb[FMT_ULONG];
    nb[fmt_ulong(nb,sa.len)]=0;
    warning(0,"debug: one_profile got ",nb," bytes",0);
  }
  if (print_one_profile(&sa) || o_all)
    return 1;
  return 0;

}

static void
do_profiles(void)
{
  int first=1;
  while (1) {
    if (0==one_profile(first))
      break;
    first=0;
    if (o_pressure_at_begin>0 || o_pressure_at_end>0)
      warning(0, "ignoring --start-pressure and --end-pressure values "
	"for the following dives",0,0,0);
      o_pressure_at_begin=0;
      o_pressure_at_end=0;
  }
}


static uogetopt2 opts[]={
{'1',"to-stdout", uogo_flag,UOGO_NOARG,&o_to_stdout,1,
/* 12345678901234567890123456789012345678901234567890 */
  "Print all profiles to stdout.",
  "In this mode mosquitoget will not stop until all "
  "profiles have been read.",0},
{'a',"all",    uogo_flag,UOGO_NOARG,&o_all,   1,
  "Get all profiles.",
  "mosquitoget will not stop when it detects what a "
  "dive profile already exists in the target "
  "directory, but will skip this profile.",0},
{'b',"bottle-volume", do_volume_opt,0, &o_bottle_volume,0,
  "Bottle volume in liter.",
  "This really is the volume, not the result of "
  "the multiplucation of pressure and volume. In other "
  "words: 'volume' is the amount of water you could "
  "fill in, not the volume the air had outside of the "
  "cylinder.","VOLUME"},
{'B',"buddies", uogo_string,0, &o_buddies,1,
  "Include NAME[S] into dive information.",
/* 12345678901234567890123456789012345678901234567890 */
  "By default the buddy information will be empty","NAME[S]"},
{'c',"config-only", uogo_flag,UOGO_NOARG, &o_conf,1,
  "Don't read profiles, print configuration to stdout.",0,0},
{'C',"comment", uogo_string,0, &o_comment,1,
  "Include COMMENT into dive information.",
  "By default there will be not comment. "
  "Only the last comment given on the command line is "
  "honored. Edit the dive profile file to include "
  "additional comments.","COMMENT"},
{'d',"device", uogo_string,0, &dev_name,0,
  "Talk to the device computer attached to DEVICE.",
  "The default device name is " DEFAULT_DEVICE ".","DEVICE"},
{'D',"dir", uogo_string,0, &o_dir,0,
  "Save profiles to DIR.",
  "The default is to save to the current directory."
  ,"DEV"},
{'e',"end-pressure", do_pressure_opt,0, &o_pressure_at_end,1,
  "Bottle pressure at end of dive.",
  "Append the unit to the number as in '750psi', though "
  "you would not need to append 'bar', which is the default unit. "
  "This will only be used for the latest dive read.\n"
  "Note that this option will be ignored if the "
  "dive profile contains information about the "
  "pressure.","PRESSURE"},
{'E',"equipment", uogo_string,0, &o_equipment,1,
  "Include EQUIPMENT into dive information.",
  "By default the equipment information will be empty.", "EQUIPMENT"},
{'m',"get-mem", callback_mem,0, 0,0,
  "Read NN bytes of memory starting at position XX.",
  "Both values are to be given in decimal numbers. "
  "The output will be sent to the standard output.","XX:NN"},
{'p',"profiles", uogo_flag,UOGO_NOARG, &o_profiles,1,
  "Get profiles (this is the default operation).",
/* 12345678901234567890123456789012345678901234567890 */
  "The program will create one file for each dive. By "
  "default this file will contain the configuration "
  "memory of the computer. Use the --without-config "
  "option to turn this off.\n"
  "The program stops when it detects another file "
  "with the same name, unless the --all option is "
  "used, in which case it not overwrite older "
  "files, but skip that dive.\n"
  "Use the --dir option to save files into another "
  "directory." ,0},
{'P',"place", uogo_string,0, &o_place,1,
  "Include PLACE into dive information.",
  "By default the place information will be empty." ,"PLACE"},
{'s',"start-pressure", do_pressure_opt,0, &o_pressure_at_begin,1,
  "Bottle pressure at start of dive.",
  "This is very much like --end-pressure.","PRESSURE"},
{'v',"verbose", uogo_flag,UOGO_NOARG, &o_verbose,1,
  "Be verbose.",0,0},
{  0,"examples",uogo_print_help,UOGO_NOARG|UOGO_EXIT|UOGO_NOLHD,
  0,0,
  "Usage examples.",
  "Read the latest dive profiles through /dev/ttyS9:\n"
  "  mosquitoget --device /dev/ttyS9\n"
  "Get all profiles not already found in /tmp:\n"
  "  mosquitoget --device=/dev/XXXX --dir /tmp --all\n"
  "Show the first 128 bytes of memory:\n"
  "  mosquitoget --device=/dev/XXXX --get-mem 0:128\n"
  "Get new profiles setting cylinder information:\n"
  "  mosquitoget --device=/dev/XXXX --bottle-volume=10liter \\\n"
  "  --start-pressure=205bar --end-pressure=75bar\n"
  "One could leave out the units in this case, since liter and bar "
  "are the defaults.\n"
  ,0},
{  0,"limitations",uogo_print_help,UOGO_NOARG|UOGO_EXIT|UOGO_NOLHD,
  0,0,
  "Known limitations.",
  "mosquitoget has not been tested against an original suunto interface, "
  "and has not been tested against other suunto dive computers. A few "
  "of them may work."
  ,0},

{  0,"see-also",uogo_print_help,UOGO_NOARG|UOGO_EXIT|UOGO_NOLHD,
  0,0,
  "Where to find related information.",
  "mosquitograph(1) turns dive profiles into pictures.\n"
  "dive-profile(5) describes the dive profile data format.\n"
  "make-dive-profile(1) creates dive profiles.\n"
  COMMON_RELATED_INFO,0},
{0,"",uogo_include,0,general_opts, 0,0,0,0},
{0,0,0,0,0,0,0,0,0}
};
uogetopt_env optenv={
  0,
  PACKAGE,
  VERSION,
  "Usage: mosquitoget [options]",

  "Read dive profiles from a suunto mosquito dive computer.",
"mosquitoget reads dive profiles and / or configuration memory from suunto\n"
"mosquito dive computers. In the default operation mode mosquitoget saves\n"
"each dive profile into one file, stopping when it discovers that a file\n"
"(dive profile) already exists.\n"
"The dive profile files may be changed using text editors, for example to\n"
"add comments.",
  COMMON_BUGREPORT_INFO,
  1,1,
  0,0,
  uogetopt_out,
  opts
};
int 
main(int argc, char **argv)
{	
  uogetopt2 *o;
  flag_bailout_fatal_begin=3;
  bailout_progname(argv[0]);
  o=uogetopt_join(opts);
  if (!o) oom();
  optenv.opts=o;
  optenv.program=flag_bailout_log_name;
  uogetopt_parse (&optenv,&argc,argv);

  dev_open();
  dev_test();
  if (!o_conf && !o_mem.len && !o_profiles)
    o_profiles=1;

  if (o_conf || o_profiles)
    get_config_memory();
  if (o_mem.len) 
    do_mem();
  if (o_conf)
    do_conf();
  if (o_profiles)
    do_profiles();

  my_DTR_off();
  if (-1==tcsetattr(dev_fd, TCSANOW, &tios_orig))
    warning(errno,"failed to restore terminal attributes",0,0,0);
  return 0;
}
