#include "buffer.h"
#include "getln.h"
#include "bailout.h"
#include "error.h"
#include "str.h"
#include "fmt.h"
#include "open.h"
#include "buffer.h"
#include "fmt.h"
#include "scan.h"
#include "uogetopt.h"
#include "stralloc.h"
#include "divedata.h"
#include "mosquitotools.h"
#include "out.h"
#include "attributes.h"
#include "scan__float.h"
#include <unistd.h>
#include <stdio.h>


static float length_conversion_factor=1;;

static unsigned long lineno=0;
static int         o_backward=0;
static const char *o_lang="german";
static const char *o_paper="a4paper";
static uogetopt2 opts[]={
  {'I',"imperial",uogo_flag,UOGO_NOARG,&o_backward,1,
    "Use units not based on international standards.",
    "For use in backwards countries: Foot, Fahrenheit."
    ,0},
  {'L',"language",uogo_string,0,&o_lang,0,
    "Language to use for LaTeX.",
    "The default is german."
    ,"LANG"},
  {'P',"paper",uogo_string,0,&o_paper,0,
    "Paper size option for LaTeX.",
    "Possible values include "
    "a4paper (default), a5paper, b5paper, letterpaper, legalpaper, "
    "executivepaper."
    ,"PAPERSIZE"},
{  0,"examples",uogo_print_help,UOGO_NOARG|UOGO_EXIT|UOGO_NOLHD,
  0,0,
    "Usage examples.",
    "Show this years dives on the screen:\n"
    "  mosquitolog 2003-0*.dive >2003-dives.tex\n"
    "  latex 2003-dives.tex && xdvi 2003-dives.dvi\n"
    "Show this years dives on the screen, using postscript:\n"
    "  mosquitolog 2003-0*.dive >2003-dives.tex\n"
    "  latex 2003-dives.tex\n"
    "  dvips -o 2003-dives.ps 2003-dives.dvi\n"
    "  gv 2003-dives.ps\n"
  ,0},

  {  0,"see-also",uogo_print_help,UOGO_NOARG|UOGO_EXIT|UOGO_NOLHD,
    0,0,
    "Where to find related information.",
    "mosquitoget(1) reads dive profiles from suunto mosquito dive computers.\n"
    "dive-profile(5) describes the dive profile data format.\n"
    COMMON_RELATED_INFO,0},

  {0,"",uogo_include,0,general_opts, 0,0,0,0},
  {0,0,0,0,0,0,0,0,0}
};

static void fill(const char *s, unsigned int i)
{
  unsigned int l=str_len(s);
  while (l++<i) 
    outs(" ");
  outs(s);
}
static float get_temp(const char *s)
{
  unsigned int l;
  float x;
  l=scan__float(s,&x);
  if (!l||s[l])
    xbailout(100,0,"cannot parse temperature ",s,0,0);
  if (o_backward) 
    return (9/5.0)*x+32;
  return x;
}

static void 
outtime(unsigned long seconds)
{
  char nb[FMT_ULONG];
  if (seconds>=600) {
    nb[fmt_ulong(nb,seconds/60)]=0;
    fill(nb,4); /* dive time */
    return;
  }
  outb(nb,fmt_ulong(nb,seconds/60));
  outs(":");
  nb[fmt_ulong(nb,seconds%60)]=0;
  if (!nb[0]) outs("00");
  else if (!nb[1]) outs("0");
  outs(nb);
}
/*                      12345678  */
static const char *sep="&";

static void 
do_header(void)
{
  outs("\\documentclass["); outs(o_paper); outs(",draft]{report}\n");
  outs("\\usepackage{umlaut}\n"); /* XXX this is likely to be not right */
  outs("\\usepackage{rotating}\n");
  outs("\\usepackage{lscape}\n");
  outs("\\usepackage{"); outs(o_lang); outs("}\n");
  outs("\\usepackage{longtable}\n");
  outs("\\pagestyle{empty}\n");
  outs("\\usepackage["); outs(o_lang); outs("]{babel}\n");
  outs("\\begin{document}\n");
  outs("\\scriptsize\n");
  outs("\\begin{landscape}\n");
  outs("\\setlength{\\tabcolsep}{0.5mm}\n");
  outs("\\begin{longtable}{|r|l|r|r|r|r|l|p{4cm}|p{4cm}|p{4cm}|}\\hline\n");
  outs("12345&8888-88-88T88:88:88&1234&9999&999.9&999.9 &999.9 - 999.9 - 999.9&"
       "Place&Buddies&Comments\\kill\n");
  outs("Nr.&Startdate&RNr&Length&Depth&Avg. &Temp. ");
  if (o_backward) outs("Fahrenheit"); else outs("Celsius"); 
  outs("&Place&Buddies&Comments\\\\\n");
  outs("   &and -time&   &Minutes&");
  if (o_backward) outs("Feet"); else outs("Meter"); 
  outs("&Depth&Start-Depth-End&"
       "     &       &\\\\\n");
  outs("\\hline\n");
  outs("\\endhead\n");
}
static void callback(struct dive *d)
{
  const char *s;
  float x;
  if (!lineno++)
    do_header();

  s=gv(&d->h_dive,"total-dive-number");
  if (s)
    outs(s);
  outs(sep);

  s=gv(&d->h_dive,"dive-start-date");
  if (!s) s="unknown";
  outs(s);
  s=gv(&d->h_dive,"dive-no");
  if (!s) s="unknown";
  outs(sep);
  fill(s,3);

  outs(sep);
  outtime(d->totalseconds);
  outs(sep);
  outf(d->maxdepth/100.0*length_conversion_factor,1); /* max depth */

  outs(sep);
  outf(d->avgdepth/100.0*length_conversion_factor,1); /* avg depth */
  outs(sep);

#define DOTEMP(tag) \
  s=gv(&d->h_dive,tag); \
  if (!s) outs("??"); else { x=get_temp(s); outf(x,1); }
  
  DOTEMP("temperature-at-begin");
  outs(" - ");
  DOTEMP("temperature-at-max-depth");
  outs(" - ");
  DOTEMP("temperature-at-end");
  outs(sep);

  s=gv(&d->h_dive,"place");
  if (!s) s="";
  outs(s);
  outs(sep);

  s=gv(&d->h_dive,"buddies");
  if (!s) s="";
  outs(s);
  outs(sep);
  {
    static stralloc sa;
    unsigned int i;
    for (i=0;;i++) {
      if (!stralloc_copys(&sa,"comment")) oom();
      if (!stralloc_catuint0(&sa,i,0)) oom();
      if (!stralloc_0(&sa)) oom();
      s=gv(&d->h_dive,sa.s);
      if (!s) break;
      if (i)
	outs("\\newline\n");
      outs(s);
    }
  }

  outs("\\\\\n");
  outs("\\hline\n");
  outflush();
}

static uogetopt_env optenv={
  0,PACKAGE,VERSION,
    "usage: mosquitolog FILE [...]\n",
    "Create table showing dive data summary",
"mosquitolog creates a LaTeX file with a summary of the dive profile FILEs\n"
"given on the command line. ",
      COMMON_BUGREPORT_INFO,
      2,0,
      0,0,
      uogetopt_out,
      opts
};


int 
main(int argc, char **argv)
{
  unsigned int i;
  uogetopt2 *o;
  flag_bailout_fatal_begin=3;
  bailout_progname(argv[0]);
  optenv.program=flag_bailout_log_name;

  o=uogetopt_join(opts);
  if (!o) oom();
  optenv.opts=o;

  uogetopt_parse (&optenv, &argc, argv);
  if (o_backward)
    length_conversion_factor=1/FOOT;

  for (i=1;argv[i];i++)
    format_handler(argv[i],callback);
  if (lineno) {
    outs("\\end{longtable}\n");
    outs("\\end{landscape}\n");
    outs("\\end{document}\n");
  }
  outflush();
  return 0;
}
